package RDF::Source::DBI::Model;
use strict;
use DBI;
use Carp;
use Data::Dumper;
use Time::HiRes qw( time );

#################################
#
#  Private methods
#
#################################

sub init
{
    my( $self, $connect, $name, $passwd ) = @_;

    $connect or croak "Connection string missing";
    $name   ||= "";
    $passwd ||= "";
    my $dbi_options ||= 
    {
#	RaiseError => 1,
    };

    $self->{'cache'} = {}; #cache->{'node'}{'uri'} ( thing, method )

    $self->dbh( DBI->connect( $connect, $name, $passwd, $dbi_options ) );
}


sub DESTROY
{
    my( $self ) = @_;
    if( defined $self->dbh )
    {
	$self->dbh->disconnect;
    }
}

sub dbh
{
    my( $self, $conn ) = @_;
    $self->{'source'}{'dbh'} = $conn if defined $conn;
    return $self->{'source'}{'dbh'};
}


sub get_node_by_id
{
    my( $self, $id ) = @_;

    my $res = $self->{'cache'}{'nodeid'}{$id};
    unless( $res )
    {
	$res = $self->new_node;
	$res->init_by_id( $id );
	$self->{'cache'}{'nodeid'}{$id} = $res;
    }
    return $res;
}

sub get_arc_by_id
{
    my( $self, $id ) = @_;
    
#    my $sta = $self->{'cache'}{'arcid'}{$id};
#    unless( $sta )
#    {
	my $sta = $self->new_arc;
	$sta->init_by_id( $id );
#	$self->{'cache'}{'arcid'}{$id} = $sta;
#    }
    return $sta;
}


sub get_uri_by_id
{
    my( $self, $id ) = @_;

#	print join("", "\n[[\t model \t get_uri_by_id \t", $self->id, "\t", $id, "\t]]\n");

    my $sth = $self->dbh->prepare_cached(qq(select uri
					    from resources
					    where id=?
					    ));
    $sth->execute( $id );
    my( $name ) = $sth->fetchrow_array;
    $sth->finish;
    
    $name ||= $id;

    return $name;
}

sub get_id_by_uri
{
    my( $self, $res ) = @_;
    # Todo: implement aboutEachPrefix

    defined $res or confess "Undefined resource";

    my $sth = $self->dbh->prepare_cached(qq(select id 
					    from resources
					    where uri=?
					    and isprefix is null
					    ));

#	print join("", "\n[[\t model \t get_id_by_uri \t", $self->id, "\t", $res, "\t]]\n");

    $sth->execute( $res );
    my( $id ) = $sth->fetchrow_array;
    $sth->finish;

    if( not defined $id )
    {
	my $ns_local = $self->ns_local;
	if( $res =~ /^$ns_local(\d+)$/ )
	{
	    $id = $1;
	}
	else
	{
	    confess "Can not find '$res' in this database";
	}
    }

    return $id;
}



#################################
#
#  Public methods
#
#################################

sub get_node_by_uri
{
    my( $self, $uri ) = @_;

    my $node = $self->{'cache'}{'node'}{$uri};
    unless( $node )
    {
	$node = $self->new_node;
	$node->init_by_uri( $uri ) or return undef;  ## Change in 0.13
	$self->{'cache'}{'node'}{$uri} = $node;
    }
    return $node;
}

sub get_arc_by_uri
{
    my( $self, $uri ) = @_;

    my $arc = $self->{'cache'}{'arc'}{$uri};
    unless( $arc )
    {
	$arc = $self->new_arc;
	$arc->init_by_uri( $uri );
	$self->{'cache'}{'arc'}{$uri} = $arc;
    }
    return $arc;
}

sub find_arcs
{
    my( $self, $pred, $subj, $obj ) = @_;

    my $arclist = $self->{'cache'}{'find_arcs'}{$pred->uri}{$subj->uri}{$obj->uri};
    unless( $arclist )
    {
	my @where = ();
	my @values = ();
	if( $pred )
	{
	    push @where, "pred = ?";
	    push @values, $pred->id;
	}
	if( $subj )
	{
	    push @where, "subj = ?";
	    push @values, $subj->id;
	}
	if( $obj )
	{
	    push @where, "obj = ?";
	    push @values, $obj->id;
	}
	
	unless( @where )
	{
	    die "find_arcs must have at least one of pred, subj, obj";
	}
	
	my $where = join " and ", @where;
	my $sql = "select id from statements where $where";

#	print join("", "\n[[\t model \t find_arcs \t", $self->id, "\t]]\n");

	my $sth = $self->dbh->prepare_cached( $sql );
	$arclist = $self->dbh->selectcol_arrayref($sth,{}, @values);
	$self->{'cache'}{'find_arcs'}{$pred->uri}{$subj->uri}{$obj->uri} = $arclist;
    }
    return $arclist; ## Returns statement ids?!?
}

sub arc_exist
{
    my( $self, $pred, $subj, $obj ) = @_;

    my $sth = $self->dbh->prepare_cached(qq( select id
					     from statements
					     where pred = ?
					     and subj = ?
					     and obj = ?
					     ));

#	print join('\t', '\n[[', 'model', 'find_arcs', $pred->id, $subj->id, $obj->id, ']]\n');

    $sth->execute( $pred->id, $subj->id, $obj->id );
    my( $id ) = $sth->fetchrow_array;
    $sth->finish;

    return 0 unless $id;

    return 1;
}

sub node_exist_with_uri
{
    my( $self, $uri ) = @_;

    my $node = $self->{'cache'}{'node'}{$uri};
    unless( $node )
    {
	my $sth = $self->dbh->prepare_cached(qq(select id 
						from resources 
						where uri=?
						));
	
#	print join('\t', '\n[[', 'model', 'node_exist_with_uri', $self->id, $uri, ']]\n');

	$sth->execute( $uri );
	my( $id ) = $sth->fetchrow_array;
	$sth->finish;

	if( not $id )
	{
	    my $ns_local = $self->ns_local;
	    $uri =~ /^$ns_local(\d+)$/ or return 0;
	}
	
    	$node = $self->new_node;
	$node->init_by_uri( $uri );
	$self->{'cache'}{'node'}{$uri} = $node;
    }
    
    return 1;
}

sub add_node
{
    my( $self, $uri, $value, $lang ) = @_;

    if( $uri )
    {
	$self->node_exist_with_uri( $uri )
	    and die "$uri already exist";
    }
    my $id = $self->nextval('internal_id');

    if( $uri || $value || $lang )
    {
	$uri ||= undef; # $self->ns_local.$id;

	my $sth = $self->dbh->prepare_cached(qq( insert into resources
						 ( id, uri, value, lang )
						 values 
						 ( ?, ?, ?, ? )
						 ));
	$sth->execute( $id, $uri, $value, $lang )
	    or confess( $sth->errstr );
    }

    my $res = $self->new_node;
    $res->init_by_id( $id );
    return $res;
}

sub add_arc
{
    my( $self, $pred, $subj, $obj, $fact ) = @_;

    if( defined $fact ) # differ between undef and 0
    {
	$fact = ( $fact ? 1 : undef );
    }
    else
    {
	$fact = 1;
    }

    if( $self->arc_exist( $pred, $subj, $obj ) )
    {
	die sprintf("Statement %s, %s, %s already exist",
		    $pred->id, $subj->id, $obj->id);
    }

    my $id = $self->nextval('internal_id');
    my $sth = $self->dbh->prepare_cached(qq( insert into statements
					     ( id, pred, subj, obj, fact )
					     values 
					     ( ?, ?, ?, ?, ? )
					     ));
    $sth->execute( $id, $pred->id, $subj->id, $obj->id, $fact )
	or confess($self->dbh->errstr);

    my $sta = $self->new_arc;
    $sta->init_by_id( $id );
    return $sta;
}

sub nextval
{
    my( $self, $seq ) = @_;

    my $sth = $self->dbh->prepare_cached( "select nextval(?)" );
    $sth->execute( $seq );
    my( $id ) = $sth->fetchrow_array;
    $sth->finish;

    $id or die "Failed to get nextval";

    return $id;
}

sub get_ContainerMembershipProperty
{
    my( $self, $nr ) = @_;
    
    my $uri = $self->ns_rdf.'_'.$nr;
    my $prop = $self->get_node_by_uri( $uri );

    return $prop if $prop;

    return $self->add_node( $uri );
}

sub get_ContainerMembershipProperties
{
    my( $self ) = @_;

    my $sth = $self->dbh->prepare_cached(qq(select id, uri
					    from resources
					    where uri like ?
					    ));

#	print join('\t', '\n[[', 'model', 'node_exist_with_uri', $self->id, ']]\n');

    my $nsrdf = $self->ns_rdf;
    $sth->execute( "$nsrdf\\_%" );

    my @member_list = ();
    while( my( $id, $uri ) = $sth->fetchrow_array )
    {
	next unless $uri =~ /^$nsrdf\_(\d+)$/;
	push @member_list, $self->get_node_by_id( $id );
    }
    $sth->finish;
    return \@member_list;
}


sub get_Literals
{
    my( $self ) = @_;

    my $sth = $self->dbh->prepare_cached(qq(select id
					    from resources
					    where value is not null
					    ));

#	print join('\t', '\n[[', 'model', 'get_Literals', $self->id, ']]\n');

    my @literals;
    $sth->execute;
    while( my( $id ) = $sth->fetchrow_array )
    {
	push @literals, $self->get_node_by_id( $id );
    }
    $sth->finish;
    return \@literals;
}

sub get_some_Resources
{
    my( $self ) = @_;

#    my $time = time; #

    my $res = $self->{'cache'}{'some_Resources'};
    unless( $res )
    {
	my $sth = $self->dbh->prepare_cached(qq(select id
						from resources
						where value is null
						));
	
	my @resources;
	$sth->execute;
	while( my( $id ) = $sth->fetchrow_array )
	{
	    push @resources, $self->get_node_by_id( $id );
	}
	$sth->finish;
	
	$res = \@resources;
	$self->{'cache'}{'some_Resources'} = $res;
#	warn "A\n";
    }
#    warn "B\n";
#    warn( (time-$time)."\n");
    return $res;
}



sub find_statements
{
# find_statements(
# 		{
#		    with =>
# 	 	    {
#  			pred => '#type',
# 			obj  => '#Class',
# 		    },
# 		    subj_lacking =>
# 		    {
# 			pred => '#domain',
# 		    }
#		    
# 		}
# 		);
#
#
# This is the same for both "get all properties without domain"
# and "get all classes without subClassOf".
#
# select subj from statements 
# where pred=18 and obj=48 and 
# subj not in ( select subj 
# from statements where pred=49);
# 
}





1;




