package RDF::Source::DBI::Resource;
use strict;
use Carp;
use Data::Dumper;
use IPC::Cache;

sub init
{
    my( $self ) = @_;
    $self->{'source'}{'dbh'} = $self->{'model'}->dbh;
}

sub dbh
{
    my( $self ) = @_;
    return $self->{'source'}{'dbh'};
}

sub init_by_id
{
    my( $self, $id ) = @_;

    my $sth = $self->dbh->prepare_cached(qq(select uri, value, isprefix, lang
					    from resources
					    where id=?
					    ));

    $sth->execute( $id );
    my( $uri, $value, $isprefix, $lang ) = $sth->fetchrow_array;
    $sth->finish;

#	print join("", "\n[[\t resource \t init_by_id \t", $id );

    # Could be an internal id, not contained in the DB
    
    $self->{'simple'}{'uri'}      = $uri || $self->model->ns_local.$id;
    $self->{'simple'}{'value'}    = $value;
    $self->{'simple'}{'lang'}     = $lang;

    $self->{'source'}{'id'}       = $id;
    $self->{'source'}{'isprefix'} = $isprefix;
}

sub init_by_uri
{
    my( $self, $uri ) = @_;
    
    confess "No uri specified" unless $uri;
    my $sth = $self->dbh->prepare_cached(qq(select id, value, isprefix, lang
					    from resources
					    where uri=?
					    ));

#	print join("", "\n[[\t resource \t init_by_uri \t", $uri );

    $sth->execute( $uri );
    my( $id, $value, $isprefix, $lang ) = $sth->fetchrow_array;
    $sth->finish;


    # Could be an internal uri, not contained in the DB
    if( not $id )
    {
	my $ns_local = $self->model->ns_local;
	$uri =~ /^$ns_local(\d+)$/ or return undef; ## Must be explicit or internal id
	$id = $1;
    }
   
    $self->{'simple'}{'uri'}      = $uri;
    $self->{'simple'}{'value'}    = $value;
    $self->{'simple'}{'lang'}     = $lang;

    $self->{'source'}{'id'}       = $id;
    $self->{'source'}{'isprefix'} = $isprefix;

    1;
}


##############################
#
#  Public methods
#
##############################

sub get_arcs_by_subject
{
    my( $self, $pred ) = @_;

#	print join("", "\n[[\t resource \t get_arcs_by_subject \t", $self->id );
	
    my( $sth, @values );
    if( $pred )
    {

#	print join("", "\t ", $pred->id );
	
	@values = ( $self->id, $pred->id );
	$sth = $self->dbh->prepare_cached(qq(select id
					     from statements
					     where subj = ?
					     and pred = ?
					     ));
    }
    else
    {
	@values = ( $self->id );
	$sth = $self->dbh->prepare_cached(qq(select id
					     from statements
					     where subj = ?
					     ));
    }

#	print "\t]]\n";

    my @sta_list = ();
    $sth->execute( @values );
    while( my( $id ) = $sth->fetchrow_array )
    {
	my $sta = $self->model->get_arc_by_id( $id );
	push @sta_list, $sta if $sta;  ## Was never undef before 0.16
#	warn "$id - $sta\n";
    }
    $sth->finish;

    return \@sta_list;
}

sub get_arcs_by_object
{
    my( $self, $pred ) = @_;

#	print join("", "\n[[\t resource \t get_arcs_by_object \t", $self->id );

    my( $sth, @values );
    if( $pred )
    {
	@values = ( $self->id, $pred->id );
	$sth = $self->dbh->prepare_cached(qq(select id
					     from statements
					     where obj = ?
					     and pred = ?
					     ));

#		print join("", "\t ", $pred->id );
    }
    else
    {
	@values = ( $self->id );
	$sth = $self->dbh->prepare_cached(qq(select id
					     from statements
					     where obj = ?
					     ));
    }

#	print "\t]]\n";

    my @sta_list = ();
    $sth->execute( @values );
    while( my( $id ) = $sth->fetchrow_array )
    {
	my $sta = $self->model->get_arc_by_id( $id );
	push @sta_list, $sta if $sta;  ## Was never undef before 0.16
    }
    $sth->finish;

    return \@sta_list;
}

sub get_objects
{
    my( $self, $pred ) = @_;

    $pred or confess "Predicate missing for ".$self->desig;

    my $subj_id = $self->id;
    my $pred_id = $pred->id;
    die unless defined $pred_id and defined $subj_id;

    my $objlist = $self->{'cache'}{'objects'}{$self->uri}{$pred->uri};

    unless( $objlist )
    {
#	print join("", "\n[[\t resource \t get_objects \t$subj_id\t$pred_id\t]]\n");
	
	
	my $sth = $self->dbh->prepare_cached(qq(select obj 
						from statements
						where subj=?
						and pred=?
						and fact is not null
						));
	$sth->execute( $subj_id,  $pred_id );
	my @obj_list = ();
	while( my( $id ) = $sth->fetchrow_array )
	{
#	    warn "id $id\n";
	    push @obj_list, $self->model->get_node_by_id( $id );
	}
	$sth->finish;
	
	$objlist = \@obj_list;
	$self->{'cache'}{'objects'}{$self->uri}{$pred->uri} = $objlist;
    }
    
#	print "\t-->\n";
    
#    warn join "-", $#$objlist, @$objlist, "\n"; #####
    return $objlist;
}

sub get_subjects
{
    my( $self, $pred ) = @_;
    $pred or confess "Predicate missing";

    my $subjlist = $self->{'cache'}{'subjects'}{$self->uri}{$pred->uri};
    unless( $subjlist )
    {
#		print join("", "\n[[\t resource \t get_subjects \t", $self->id, "\t", $pred->id, "\t]]\n");

	my $pred_id  = $pred->id;
	my $obj_id   = $self->id;
	die unless defined $pred_id and defined $obj_id;
	
	my $sth = $self->dbh->prepare_cached(qq(select subj 
						from statements
						where pred=?
						and obj=?
						and fact is not null
						));
	$sth->execute( $pred_id, $obj_id );
		my @subj_list = ();
	while( my( $id ) = $sth->fetchrow_array )
	{
	    push @subj_list, $self->model->get_node_by_id( $id );
	}
	$sth->finish;

	$subjlist = \@subj_list;
	$self->{'cache'}{'subjects'}{$self->uri}{$pred->uri} = $subjlist;
    }

    return $subjlist;
}

sub has_property
{
    my( $self, $pred ) = @_;

    my $sth = $self->dbh->prepare_cached(qq(select id 
					    from statements 
					    where subj = ?
					    and pred = ?
					    ));

#	print join("", "\n[[\t resource \t has_property \t", $self->id, "\t", $pred->id, "\t]]\n");
    
    $sth->execute( $self->id, $pred->id );
    my( $id ) = $sth->fetchrow_array;
    $sth->finish;

    return $id ? 1 : undef;
}


sub get_membership_objects
{
    my( $self ) = @_;

    my $nsrdf = $self->model->ns_rdf;

    my $sth = $self->dbh->prepare_cached(qq(select uri, obj 
					    from statements, resources
					    where subj=?
					    and pred=resources.id
					    and uri like ?
					    and fact is not null
					    ));

#	print join("", "\n[[\t resource \t get_membership_objects \t", $self->id, "\t]]\n");

    $sth->execute( $self->id, "$nsrdf\\_%" );
    my @obj_list = ();
    while( my( $uri, $obj ) = $sth->fetchrow_array )
    {
	next unless $uri =~ /^$nsrdf\_(\d+)$/;
	$obj_list[ $1 ] = $self->model->get_node_by_id( $obj );
    }
    $sth->finish;

    return [ grep { $_ if defined } @obj_list ]; # Do not return empty slots
}

sub get_membership_arcs
{
    my( $self ) = @_;

    my $nsrdf = $self->model->ns_rdf;

    my $sth = $self->dbh->prepare_cached(qq(select uri, obj, statements.id
					    from statements, resources
					    where subj=?
					    and pred=resources.id
					    and uri like ?
					    and fact is not null
					    ));

#	print join("", "\n[[\t resource \t get_membership_arcs \t", $self->id, "\t]]\n");


    $sth->execute( $self->id, "$nsrdf\\_%" );
    my @arc_list = ();
    while( my( $uri, $obj, $id ) = $sth->fetchrow_array )
    {
	next unless $uri =~ /^$nsrdf\_(\d+)$/;
	$arc_list[ $1 ] = $self->model->get_arc_by_id( $id );
    }
    $sth->finish;
    return [ grep { $_ if defined } @arc_list ]; # Do not return empty slots
}

sub delete
{
    my( $self ) = @_;

    my $sth = $self->dbh->prepare_cached(qq(delete from resources 
					    where id = ?
					    ));

    $sth->execute( $self->id ) or die $sth->errstr;

    undef $self;
}

sub store
{
    my( $self ) = @_;

    my $sth = $self->dbh->prepare_cached(qq(update resources set
					    uri=?, lang=?, value=?
					    where id=?
					    ));
    
    $sth->execute( $self->uri, $self->lang, $self->value, $self->id )
	or die $sth->errstr;
}

sub id       { $_[0]->{'source'}{'id'} }
sub isprefix { $_[0]->{'source'}{'isprefix'} }


1;





